from decoupler._Method import _show_methods
from decoupler.mt._aucell import aucell
from decoupler.mt._gsea import gsea
from decoupler.mt._gsva import gsva
from decoupler.mt._mdt import mdt
from decoupler.mt._mlm import mlm
from decoupler.mt._ora import ora
from decoupler.mt._udt import udt
from decoupler.mt._ulm import ulm
from decoupler.mt._viper import viper
from decoupler.mt._waggr import waggr
from decoupler.mt._zscore import zscore

_methods = [
    aucell,
    gsea,
    gsva,
    mdt,
    mlm,
    ora,
    udt,
    ulm,
    viper,
    waggr,
    zscore,
]
