from decoupler.ds._bulk import hsctgfb, knocktf
from decoupler.ds._scell import pbmc3k, covid5k, erygast1k
from decoupler.ds._spatial import msvisium
from decoupler.ds._toy import toy, toy_bench
from decoupler.ds._utils import ensmbl_to_symbol
